/* -*- Mode: c; c-basic-offset: 2 -*-
 *
 * (c) 2007 Satis Superque Merce B.V. All rights reserved.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * rdf_storage_b.c - RDF Storage using B implementation
 */

#ifdef HAVE_CONFIG_H
#include <rdf_config.h>
#endif

#ifdef WIN32
#include <win32_rdf_config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <time.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h> /* for abort() as used in errors */
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>

#include <redland.h>
#include <rdf_storage.h>
#include <rdf_digest.h>
#include <rdf_types.h>

#include <b.h>

typedef struct
{
  librdf_storage *storage;

  b_t *db;
  int db_cache;

  unsigned char *path;

} librdf_storage_b_context;

typedef struct {
  librdf_storage *storage;
  librdf_storage_b_context* b_context;

  int finished;

  librdf_statement *statement;
  librdf_node* context;

  b_iterator_triple_t *iterator;
} librdf_storage_b_stream_context;

typedef struct {
  librdf_storage *storage;
  librdf_storage_b_context* b_context;

  int finished;
  b_iterator_type_t *iterator;
  unsigned char *what;
} librdf_storage_b_iterator_context;

static b_triple_t * librdf_storage_b_triple(librdf_storage* storage,
                                            librdf_node *subject_node, 
		                            librdf_node *property_node,
                                            librdf_node *object_node,
					    librdf_node *context_node);
static void librdf_storage_b_triple_free(unsigned char **values);

static int librdf_storage_b_stream_end_of_stream(void* context);
static int librdf_storage_b_stream_next_statement(void* context);
static void librdf_storage_b_stream_finished(void* context);
static void* librdf_storage_b_stream_get_statement(void* context, int flags);

static int librdf_storage_b_iterator_is_end(void* iterator);
static int librdf_storage_b_iterator_next_method(void* iterator);
static void* librdf_storage_b_iterator_get_method(void* iterator, int);
static void librdf_storage_b_iterator_finished(void* iterator);

static int
librdf_storage_b_init(librdf_storage* storage, char *name, librdf_hash* options)
{
  librdf_storage_b_context *context=(librdf_storage_b_context*)storage->context;
  unsigned char *path;
  int len;

  if(!name)
    return 1;

  context->storage = storage;

  if(options && (context->db_cache=(int)librdf_hash_get_as_long(options, "cache"))<0)
    context->db_cache=0;

  len=strlen(name);

  if(!(path=(unsigned char*)LIBRDF_MALLOC(cstring, len+1)))
    return 1;

  strncpy((char *)path, name, len+1);
  context->path=path;

  if(options)
    librdf_free_hash(options);

  return 0;
}

static void
librdf_storage_b_terminate(librdf_storage* storage)
{
  librdf_storage_b_context *context=(librdf_storage_b_context*)storage->context;

  if(context->path)
    LIBRDF_FREE(cstring, context->path);
}

static int
librdf_storage_b_open(librdf_storage* storage, librdf_model* model)
{
  librdf_storage_b_context *context=(librdf_storage_b_context*)storage->context;
  b_error_t error;

  if((error=b_new(&context->db, context->path))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, NULL,
               "B database %s open failed - %s", context->path,
               b_strerror(error));
    return 1;
  }

  if(context->db_cache)
    b_set_max_cache(context->db, context->db_cache);

  return 0;
}

static int
librdf_storage_b_close(librdf_storage* storage)
{
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  b_error_t error;

  if(context->db) {
    if((error=b_destroy(context->db))!=B_OK) {
      librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
                 NULL, "B database %s close failed - %s", context->path,
                 b_strerror(error));
    }
    context->db=NULL;
  }

  return 0;
}

static int
librdf_storage_b_size(librdf_storage* storage)
{
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  b_uint64 count=0;
  b_error_t error;

  if((error=b_count_triple(context->db, &count))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B database %s get size failed - %s", context->path,
               b_strerror(error));
  }

  return (int)count;
}

static int
librdf_storage_b_context_add_statement(librdf_storage* storage,
                                       librdf_node* context_node,
                                       librdf_statement* statement)
{
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  b_triple_t *triple;
  b_error_t error;

  if(!(triple= librdf_storage_b_triple(storage, 
                                       librdf_statement_get_subject(statement), 
                                       librdf_statement_get_predicate(statement),
                                       librdf_statement_get_object(statement),
                                       context_node)))
    return 1;

  if((error=b_add_triple(context->db, triple))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B database %s add triple failed - %s", context->path,
               b_strerror(error));

    b_triple_destroy(triple);
    return 1;
  }

  b_triple_destroy(triple);
  return 0;
}

static int
librdf_storage_b_add_statement(librdf_storage* storage,
                                    librdf_statement* statement)
{
  return librdf_storage_b_context_add_statement(storage, NULL, statement);
}

static unsigned char **
librdf_storage_b_triple_array(librdf_storage *storage,
                              librdf_node *subject_node,
			      librdf_node *property_node,
                              librdf_node *object_node,
			      librdf_node *context_node,
			      unsigned int *values_len)
{
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  unsigned char **values;
  librdf_uri *uri;

  if(!(values=calloc(9, sizeof(unsigned char *))))
    return NULL;

  memset(values_len, 0, sizeof(unsigned int)*9);

#define B_FIND_NODE( x , a , b , c , d , e) \
  if(x) { \
    switch(librdf_node_get_type(x)) { \
      case LIBRDF_NODE_TYPE_RESOURCE: \
        values[a]=librdf_uri_as_counted_string(librdf_node_get_uri(x), \
                                               &values_len[a]); \
        break; \
      \
      case LIBRDF_NODE_TYPE_BLANK: \
        values[b]=librdf_node_get_blank_identifier(x); \
        break; \
      \
      case LIBRDF_NODE_TYPE_LITERAL: \
        values[c]=librdf_node_get_literal_value_as_counted_string(x, \
                                                    &values_len[c]); \
        if ((uri=librdf_node_get_literal_value_datatype_uri(x))) \
          values[d]=(unsigned char *)librdf_uri_as_counted_string(uri, \
			                                     &values_len[d]); \
        values[e]=(unsigned char *)librdf_node_get_literal_value_language(x); \
        break; \
      \
      case LIBRDF_NODE_TYPE_UNKNOWN: \
        break; \
      \
    } \
  }

  B_FIND_NODE(subject_node, 0, 1, 1, 1, 1);
  B_FIND_NODE(property_node, 2, 2, 2, 2, 2);
  B_FIND_NODE(object_node, 3, 4, 5, 7, 8);
  B_FIND_NODE(context_node, 6, 6, 6, 6, 6);

  return values;
}

static b_triple_t *
librdf_storage_b_triple(librdf_storage *storage,
                        librdf_node *subject_node, librdf_node *property_node,
                        librdf_node *object_node, librdf_node *context_node)
{
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  unsigned char **values;
  unsigned int values_len[9];
  b_triple_t *triple;
  b_error_t error;

  if(!(values = librdf_storage_b_triple_array(storage, subject_node, 
				              property_node,object_node,
					      context_node, values_len)))
	  return NULL;

  if((error=b_triple_new_incomplete(&triple, values[0], values_len[0], 
				    values[1], values_len[1],
				    values[2], values_len[2],
				    values[3], values_len[3],
				    values[4], values_len[4],
				    values[5], values_len[5],
				    values[6], values_len[6],
				    values[7], values_len[7],
				    values[8], values_len[8]))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B database %s create triple failed - %s", context->path,
               b_strerror(error));
    librdf_storage_b_triple_free(values);
    return NULL;
  }

  librdf_storage_b_triple_free(values);
  return triple;
}

static void librdf_storage_b_triple_free(unsigned char **values) {
/*
  int i;
  for(i=0;i<9;i++)
    if(values[i])
      free(values[i]);
*/

  free(values);
}

static int
librdf_storage_b_context_remove_statement(librdf_storage* storage,
                                          librdf_node* context_node,
                                          librdf_statement* statement)
{
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  b_triple_t *triple;
  b_error_t error;

  if(!(triple= librdf_storage_b_triple(storage,
                                       librdf_statement_get_subject(statement), 
                                       librdf_statement_get_predicate(statement),
                                       librdf_statement_get_object(statement),
                                       context_node)))
    return 1;

  if((error=b_remove_triple(context->db, triple))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B database %s remove triple failed - %s", context->path,
               b_strerror(error));

    b_triple_destroy(triple);
    return 1;
  }

  b_triple_destroy(triple);
  return 0;
}

static int
librdf_storage_b_remove_statement(librdf_storage* storage, 
                                  librdf_statement* statement)
{
  return librdf_storage_b_context_remove_statement(storage, NULL,
                                                   statement);
}

static int
librdf_storage_b_contains_statement(librdf_storage* storage,
                                         librdf_statement* statement)
{
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  b_triple_t *triple;
  b_error_t error;
  char status;

  if(!(triple= librdf_storage_b_triple(storage,
                                       librdf_statement_get_subject(statement), 
                                       librdf_statement_get_predicate(statement),
                                       librdf_statement_get_object(statement),
                                       NULL)))
    return 1;

  if((error=b_get_triple(context->db, triple, &status))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B database %s remove triple failed - %s", context->path,
               b_strerror(error));

    b_triple_destroy(triple);
    return 1;
  }

  b_triple_destroy(triple);
  return status ? 1 : 0;
}

static librdf_stream*
librdf_storage_b_serialise(librdf_storage* storage)
{
  librdf_stream *stream;
  librdf_storage_b_stream_context *scontext;
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  b_error_t error;

  scontext=(librdf_storage_b_stream_context*)LIBRDF_CALLOC(librdf_storage_b_stream_context, 1, sizeof(librdf_storage_b_stream_context));
  scontext->storage=storage;
  scontext->b_context=context;

  librdf_storage_add_reference(scontext->storage);

  if((error=b_iterator_triple_new(context->db, &scontext->iterator, NULL))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B iterator create failed - %s", context->path,
               b_strerror(error));
    librdf_storage_b_stream_finished((void*)scontext);
    return NULL;
  }
	  
  if(!(stream=librdf_new_stream(storage->world,
                                (void*)scontext,
                                &librdf_storage_b_stream_end_of_stream,
                                &librdf_storage_b_stream_next_statement,
                                &librdf_storage_b_stream_get_statement,
                                &librdf_storage_b_stream_finished))) {
    librdf_storage_b_stream_finished((void*)scontext);
    return NULL;
  }

  return stream;
}

static int
librdf_storage_b_get_next_stream_common(librdf_storage_b_stream_context* scontext,
                                        b_iterator_triple_t *iterator,
                                        librdf_statement **statement)
{
  librdf_node *node;
  b_triple_t *next;
  b_error_t error;
 
  if((error=b_iterator_triple_step(iterator, &next))!=B_OK) {
    librdf_log(scontext->storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B database %s iterator triple step failed - %s",
               scontext->b_context->path, b_strerror(error));
    return 1;
  }

  if(!next)
    return 1;

  if(!(*statement=librdf_new_statement(scontext->storage->world)))
    return 1;

  librdf_statement_clear(*statement);

  if(next->subject_uri)
    node=librdf_new_node_from_uri_string(scontext->storage->world,
                                         next->subject_uri);
  else
    node=librdf_new_node_from_blank_identifier(scontext->storage->world,
                                               next->subject_bnode);

  librdf_statement_set_subject(*statement, node);

  node=librdf_new_node_from_uri_string(scontext->storage->world,
                                       next->property);

  librdf_statement_set_predicate(*statement, node);

  if(next->object_uri)
    node=librdf_new_node_from_uri_string(scontext->storage->world,
                                         next->object_uri);
  else if(next->object_bnode)
    node=librdf_new_node_from_blank_identifier(scontext->storage->world,
                                               next->object_bnode);
  else {
    librdf_uri *datatype;

    if(next->datatype)
      datatype=librdf_new_uri(scontext->storage->world,
                              next->datatype);
    else
      datatype=NULL;

    node=librdf_new_node_from_typed_literal(scontext->storage->world,
                                            next->object_literal,
                                            (char *)next->lang,
                                            datatype);

    if(datatype)
      librdf_free_uri(datatype);
  }

  librdf_statement_set_object(*statement, node);

  return 0;
}

static int
librdf_storage_b_stream_end_of_stream(void* context)
{
  librdf_storage_b_stream_context* scontext=(librdf_storage_b_stream_context*)context;

  if(scontext->finished)
    return 1;

  if(scontext->statement == NULL) {
    if(librdf_storage_b_get_next_stream_common(scontext, scontext->iterator, &scontext->statement))
      scontext->finished=1;
  }

  return scontext->finished;
}

static int
librdf_storage_b_stream_next_statement(void* context)
{
  librdf_storage_b_stream_context* scontext=(librdf_storage_b_stream_context*)context;

  if(scontext->finished)
    return 1;

  if(librdf_storage_b_get_next_stream_common(scontext, scontext->iterator, &scontext->statement))
    scontext->finished=1;

  return scontext->finished;
}

static void*
librdf_storage_b_stream_get_statement(void* context, int flags)
{
  librdf_storage_b_stream_context* scontext=(librdf_storage_b_stream_context*)context;

  switch(flags) {
    case LIBRDF_ITERATOR_GET_METHOD_GET_OBJECT:
      return scontext->statement;
    case LIBRDF_ITERATOR_GET_METHOD_GET_CONTEXT:
      return scontext->context;
    default:
      librdf_log(scontext->storage->world,
                 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, NULL,
                 "Unknown iterator method flag %d", flags);
      return NULL;
  }
}

static void
librdf_storage_b_stream_finished(void* context)
{
  librdf_storage_b_stream_context* scontext=(librdf_storage_b_stream_context*)context;

  if(scontext->storage)
    librdf_storage_remove_reference(scontext->storage);

  if(scontext->statement)
    librdf_free_statement(scontext->statement);

  if(scontext->context)
    librdf_free_node(scontext->context);

  if(scontext->iterator)
    b_iterator_triple_destroy(scontext->iterator);

  LIBRDF_FREE(librdf_storage_b_stream_context, scontext);
}

static librdf_stream*
librdf_storage_b_find_statements(librdf_storage* storage, librdf_statement* statement)
{
  librdf_stream *stream;
  librdf_storage_b_stream_context *scontext;
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  b_triple_t *compare;
  b_error_t error;

  scontext=(librdf_storage_b_stream_context*)LIBRDF_CALLOC(librdf_storage_b_stream_context, 1, sizeof(librdf_storage_b_stream_context));
  scontext->storage=storage;
  scontext->b_context=context;

  librdf_storage_add_reference(scontext->storage);

  if(!(compare = librdf_storage_b_triple(storage,
				         librdf_statement_get_subject(statement), 
					 librdf_statement_get_predicate(statement),
					 librdf_statement_get_object(statement), 
					 NULL))) {
    librdf_storage_b_stream_finished((void*)scontext);
    return NULL;
  }

  if((error=b_iterator_triple_new(context->db, &scontext->iterator, compare))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B iterator create failed - %s", context->path,
               b_strerror(error));

    librdf_storage_b_stream_finished((void*)scontext);
    b_triple_destroy(compare);
    return NULL;
  }
	  
  b_triple_destroy(compare);

  if(!(stream=librdf_new_stream(storage->world,
                                (void*)scontext,
                                &librdf_storage_b_stream_end_of_stream,
                                &librdf_storage_b_stream_next_statement,
                                &librdf_storage_b_stream_get_statement,
                                &librdf_storage_b_stream_finished))) {
    librdf_storage_b_stream_finished((void*)scontext);
    return NULL;
  }

  return stream;
}

static librdf_stream*
librdf_storage_b_context_serialise(librdf_storage* storage, librdf_node* context_node)
{
  librdf_stream *stream;
  librdf_storage_b_stream_context *scontext;
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  b_triple_t *compare;
  b_error_t error;

  scontext=(librdf_storage_b_stream_context*)LIBRDF_CALLOC(librdf_storage_b_stream_context, 1, sizeof(librdf_storage_b_stream_context));
  scontext->storage=storage;
  scontext->b_context=context;

  librdf_storage_add_reference(scontext->storage);

  if(!(compare = librdf_storage_b_triple (storage, NULL, NULL, NULL, context_node))) {
    librdf_storage_b_stream_finished((void*)scontext);
    return NULL;
  }

  if((error=b_iterator_triple_new(context->db, &scontext->iterator, compare))!=B_OK) {
    librdf_log(storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B iterator create failed - %s", context->path,
               b_strerror(error));

    librdf_storage_b_stream_finished((void*)scontext);
    b_triple_destroy(compare);
    return NULL;
  }
	  
  b_triple_destroy(compare);

  if(!(stream=librdf_new_stream(storage->world,
                                (void*)scontext,
                                &librdf_storage_b_stream_end_of_stream,
                                &librdf_storage_b_stream_next_statement,
                                &librdf_storage_b_stream_get_statement,
                                &librdf_storage_b_stream_finished))) {
    librdf_storage_b_stream_finished((void*)scontext);
    return NULL;
  }

  return stream;
}

static librdf_iterator*
librdf_storage_b_get_contexts(librdf_storage* storage) {
  librdf_storage_b_context* context=(librdf_storage_b_context*)storage->context;
  librdf_storage_b_iterator_context* icontext;
  librdf_iterator* iterator;
  b_error_t error;

  icontext=(librdf_storage_b_iterator_context*)LIBRDF_CALLOC(librdf_storage_b_iterator_context, 1, sizeof(librdf_storage_b_iterator_context));
  if(!icontext)
    return NULL;

  icontext->b_context=context;
  icontext->storage=storage;

  librdf_storage_add_reference(icontext->storage);

  if((error=b_iterator_type_new(context->db, B_TYPE_CONTEXT, &icontext->iterator))!=B_OK) {
    librdf_storage_b_iterator_finished((void*)icontext);
    return NULL;
  }

  if(!(iterator=librdf_new_iterator(storage->world,
                                    (void*)icontext,
                                    &librdf_storage_b_iterator_is_end,
                                    &librdf_storage_b_iterator_next_method,
                                    &librdf_storage_b_iterator_get_method,
                                    &librdf_storage_b_iterator_finished))) {
    librdf_storage_b_iterator_finished((void*)icontext);
    return NULL;
  }

  return iterator;
}

static int
librdf_storage_b_get_next_iterator_common(librdf_storage_b_iterator_context* icontext)
{
  b_error_t error;

  if(icontext->what)
    LIBRDF_FREE(cstring, icontext->what);

  if((error=b_iterator_type_step(icontext->iterator, &icontext->what, NULL))!=B_OK)  {
    librdf_log(icontext->storage->world, 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, 
               NULL, "B database %s iterator type step failed - %s",
               icontext->b_context->path, b_strerror(error));
    icontext->what=NULL;
    return 1;
  }

  if(!icontext->what)
    return 1;

  return 0;
}

static int librdf_storage_b_iterator_is_end(void* iterator)
{
  librdf_storage_b_iterator_context* icontext=(librdf_storage_b_iterator_context*)iterator;

  if(icontext->finished)
    return 1;

  if(!icontext->what)
    if(librdf_storage_b_get_next_iterator_common(icontext))
      icontext->finished=1;

  return icontext->finished;
}

static int librdf_storage_b_iterator_next_method(void* iterator)
{
  librdf_storage_b_iterator_context* icontext=(librdf_storage_b_iterator_context*)iterator;

  if(icontext->finished)
    return 1;

  if(librdf_storage_b_get_next_iterator_common(icontext))
    icontext->finished=1;

  return icontext->finished;
}

static void* librdf_storage_b_iterator_get_method(void* iterator, int flags) 
{
  librdf_storage_b_iterator_context* icontext=(librdf_storage_b_iterator_context*)iterator;

  switch(flags) {
    case LIBRDF_ITERATOR_GET_METHOD_GET_OBJECT:
      if(icontext->what)
        return librdf_new_node_from_uri_string(icontext->storage->world,
                                               icontext->what);
      else
        return NULL;

    case LIBRDF_ITERATOR_GET_METHOD_GET_KEY:
    case LIBRDF_ITERATOR_GET_METHOD_GET_VALUE:
      return NULL;

    default:
      librdf_log(icontext->storage->world,
                 0, LIBRDF_LOG_ERROR, LIBRDF_FROM_STORAGE, NULL,
                 "Unknown iterator method flag %d", flags);
    return NULL;
  }
}

static void librdf_storage_b_iterator_finished(void* iterator)
{
  librdf_storage_b_iterator_context* icontext=(librdf_storage_b_iterator_context*)iterator;

  if(icontext->storage)
    librdf_storage_remove_reference(icontext->storage);

  if(icontext->what)
    free(icontext->what);

  LIBRDF_FREE(librdf_storage_b_iterator_context, icontext);
}

static librdf_node*
librdf_storage_b_get_feature(librdf_storage* storage, librdf_uri* feature)
{
  /* librdf_storage_b_context* scontext=(librdf_storage_b_context*)storage->context; */
  unsigned char *uri_string;

  if(!feature)
    return NULL;

  if(!(uri_string=librdf_uri_as_string(feature)))
    return NULL;

  if(!strcmp((const char*)uri_string, LIBRDF_MODEL_FEATURE_CONTEXTS)) {
    return librdf_new_node_from_typed_literal(storage->world,
                                              (const unsigned char*)"1",
                                              NULL, NULL);
  }

  return NULL;
}

/* local function to register B storage functions */
static void
librdf_storage_b_register_factory(librdf_storage_factory *factory) 
{
  factory->context_length     = sizeof(librdf_storage_b_context);
  
  factory->init               = librdf_storage_b_init;
  factory->terminate          = librdf_storage_b_terminate;
  factory->open               = librdf_storage_b_open;
  factory->close              = librdf_storage_b_close;
  factory->size               = librdf_storage_b_size;
  factory->add_statement      = librdf_storage_b_add_statement;
  factory->remove_statement   = librdf_storage_b_remove_statement;
  factory->contains_statement = librdf_storage_b_contains_statement;
  factory->serialise          = librdf_storage_b_serialise;
  factory->find_statements    = librdf_storage_b_find_statements;
  factory->context_add_statement = librdf_storage_b_context_add_statement;
  factory->context_remove_statement = librdf_storage_b_context_remove_statement;
  factory->context_serialise  = librdf_storage_b_context_serialise;
  factory->get_contexts       = librdf_storage_b_get_contexts;
  factory->get_feature        = librdf_storage_b_get_feature;
}

void
librdf_init_storage_b(librdf_world *world)
{
  librdf_storage_register_factory(world, "b", "B",
                                  &librdf_storage_b_register_factory);
}

